#include "chunk.h"

Chunk::Chunk(int x, int y, int z)
    : x(x), y(y), z(z)
{
    //Everything air
    memset(blocks, 0, sizeof(blocks));

    printf("Generating chunk %d:%d:%d...\t", x, y, z);

    for(int x = 0; x < SIZE; x++)
        for(int z = 0; z < SIZE; z++)
        {
            int height = (sinf(float(x)/4) * sinf(float(z)/4)) * 16 + 1;
            for(int y = 0; y < height; y++)
                blocks[x][y][z] = z < 3 ? 1 : (height > 8 ? 2 : 3);
        }

    printf("Done!\n");
}

Chunk::~Chunk()
{
    delete[] buffer;
}

#define IN_BOUNDS(x, y, z) (x >= 0 && y >= 0 && z >= 0 && x < SIZE && y < SIZE && z < SIZE)

void Chunk::render()
{
    if(is_dirty)
    {
        printf("Updating chunk %d:%d:%d...\t\t", x, y, z);
        delete[] buffer;
        buffer_length = 0;

        //Calculate size of buffer
        for(int x = -1; x <= SIZE; x++)
            for(int y = -1; y <= SIZE; y++)
                for(int z = -1; z <= SIZE; z++)
                {
                    //If not air, sides of adjacent blocks aren't visible
                    if(IN_BOUNDS(x, y, z) && blocks[x][y][z] != 0)
                        continue;

                    if(IN_BOUNDS(x - 1, y, z) && blocks[x - 1][y][z] != 0)
                            buffer_length += 4;

                    if(IN_BOUNDS(x + 1, y, z) && blocks[x + 1][y][z] != 0)
                            buffer_length += 4;

                    if(IN_BOUNDS(x, y - 1, z) && blocks[x][y - 1][z] != 0)
                            buffer_length += 4;

                    if(IN_BOUNDS(x, y + 1, z) && blocks[x][y + 1][z] != 0)
                            buffer_length += 4;

                    if(IN_BOUNDS(x, y, z - 1) && blocks[x][y][z - 1] != 0)
                            buffer_length += 4;

                    if(IN_BOUNDS(x, y, z + 1) && blocks[x][y][z + 1] != 0)
                            buffer_length += 4;
                }

        buffer = new VERTEX[buffer_length];

        //Now, render!

        VERTEX *buffer2 = buffer;

        GLFix posX = this->x * BLOCK_SIZE - BLOCK_SIZE;

        for(int x = -1; x <= SIZE; x++, posX += BLOCK_SIZE)
        {
            GLFix posY = this->y * BLOCK_SIZE - BLOCK_SIZE;
            for(int y = -1; y <= SIZE; y++, posY += BLOCK_SIZE)
            {
                GLFix posZ = this->z * BLOCK_SIZE - BLOCK_SIZE;
                for(int z = -1; z <= SIZE; z++, posZ += BLOCK_SIZE)
                {
                    //If not air, sides of adjacent blocks aren't visible
                    if(IN_BOUNDS(x, y, z) && blocks[x][y][z] != 0)
                        continue;

                    if(IN_BOUNDS(x - 1, y, z) && blocks[x - 1][y][z] != 0)
                    {
                        GLFix myposX = posX - BLOCK_SIZE;
                        GLFix myposY = posY;
                        GLFix myposZ = posZ;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x - 1][y][z]];

                            if((x + y + z - 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &ri = block_textures[blocks[x - 1][y][z]];
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ, ri.right, ri.bottom, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ, ri.right, ri.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, ri.left, ri.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ + BLOCK_SIZE, ri.left, ri.bottom, c};
                    }

                    if(IN_BOUNDS(x + 1, y, z) && blocks[x + 1][y][z] != 0)
                    {
                        GLFix myposX = posX + BLOCK_SIZE;
                        GLFix myposY = posY;
                        GLFix myposZ = posZ;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x + 1][y][z]];

                            if((x + y + z + 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &le = block_textures[blocks[x + 1][y][z]];
                        *buffer2++ = {myposX, myposY, myposZ + BLOCK_SIZE, le.left, le.bottom, c};
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, le.left, le.top, c};
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ, le.right, le.top, c};
                        *buffer2++ = {myposX, myposY, myposZ, le.right, le.bottom, c};
                    }

                    if(IN_BOUNDS(x, y - 1, z) && blocks[x][y - 1][z] != 0)
                    {
                        GLFix myposX = posX;
                        GLFix myposY = posY - BLOCK_SIZE;
                        GLFix myposZ = posZ;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x][y - 1][z]];

                            if((x + y + z - 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &to = block_textures[blocks[x][y - 1][z]];
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ, to.left, to.bottom, c};
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, to.left, to.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, to.right, to.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ, to.right, to.bottom, c};
                    }

                    if(IN_BOUNDS(x, y + 1, z) && blocks[x][y + 1][z] != 0)
                    {
                        GLFix myposX = posX;
                        GLFix myposY = posY + BLOCK_SIZE;
                        GLFix myposZ = posZ;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x][y + 1][z]];

                            if((x + y + z + 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &bo = block_textures[blocks[x][y + 1][z]];
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ, bo.left, bo.bottom, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ + BLOCK_SIZE, bo.left, bo.top, c};
                        *buffer2++ = {myposX, myposY, myposZ + BLOCK_SIZE, bo.right, bo.top, c};
                        *buffer2++ = {myposX, myposY, myposZ, bo.right, bo.bottom, c};
                    }

                    if(IN_BOUNDS(x, y, z - 1) && blocks[x][y][z - 1] != 0)
                    {
                        GLFix myposX = posX;
                        GLFix myposY = posY;
                        GLFix myposZ = posZ - BLOCK_SIZE;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x][y][z - 1]];

                            if((x + y + z - 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &ba = block_textures[blocks[x][y][z - 1]];
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ + BLOCK_SIZE, ba.left, ba.bottom, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, ba.left, ba.top, c};
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ + BLOCK_SIZE, ba.right, ba.top, c};
                        *buffer2++ = {myposX, myposY, myposZ + BLOCK_SIZE, ba.right, ba.bottom, c};
                    }

                    if(IN_BOUNDS(x, y, z + 1) && blocks[x][y][z + 1] != 0)
                    {
                        GLFix myposX = posX;
                        GLFix myposY = posY;
                        GLFix myposZ = posZ + BLOCK_SIZE;

                        #ifndef TEXTURE_SUPPORT
                            COLOR c = block_colors[blocks[x][y][z + 1]];

                            if((x + y + z + 1) % 2 == 0)
                            {
                                RGB rgb = rgbColor(c);
                                rgb.r *= 0.8f;
                                rgb.g *= 0.8f;
                                rgb.b *= 0.8f;

                                c = colorRGB(rgb.r, rgb.g, rgb.b);
                            }
                        #else
                            COLOR c = 0;
                        #endif

                        TextureAtlasEntry &fr = block_textures[blocks[x][y][z + 1]];
                        *buffer2++ = {myposX, myposY, myposZ, fr.left, fr.bottom, c};
                        *buffer2++ = {myposX, myposY + BLOCK_SIZE, myposZ, fr.left, fr.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY + BLOCK_SIZE, myposZ, fr.right, fr.top, c};
                        *buffer2++ = {myposX + BLOCK_SIZE, myposY, myposZ, fr.right, fr.bottom, c};
                    }
                }
            }
        }

        is_dirty = false;

        printf("Done!\n");
    }

    glBegin(GL_QUADS);
    nglAddVertices(buffer, buffer_length);
    glEnd();
}
