#ifndef GL_H
#define GL_H

//nGL version 0.3

#include "fix.h"

//Either colors or textures, but not both
#define TEXTURE_SUPPORT

//Warning, this will be slower than textures!
//#define INTERPOLATE_COLORS

//#define WIREFRAME_MODE

//Deactivate after you don't get any warnings.
//Otherwise it may corrupt random memory and crash.
//#define SAFE_MODE

//Whether to render colors or draw the z_buffer (depth buffer) onto the screen
//White = near, black = far
//#define DRAW_Z_BUFFER

#if defined(TEXTURE_SUPPORT) && defined(INTERPOLATE_COLORS)
#error "Colors and textures cannot be used simultaneously!"
#endif

//Works best with 300
#define NEAR_PLANE 300
#define CLIP_PLANE 150

#define MATRIX_STACK_SIZE 5

//These values are used to calculate offsets into the buffer.
//If you want something like FBOs, make them variables and set them accordingly.
//Watch out for different buffer sizes!
#define WIDTH 320
#define HEIGHT 240

//Change if you want to, but it may not work.
typedef Fix<9, int_fast32_t> GLFix;

/* Column vectors and matrices:
 * [ [0][0] [0][1] [0][2] [0][3] ]   [x]
 * [ [1][0] [1][1] [1][2] [1][3] ] * [y]
 * [ [2][0] [2][1] [2][2] [2][3] ]   [z]
 * [ [3][0] [3][1] [3][2] [3][3] ]   [1] (not used anywhere)
 */

typedef uint16_t COLOR;
typedef struct VERTEX { GLFix x; GLFix y; GLFix z; GLFix u; GLFix v; COLOR c; } VERTEX;

//Width and height should be smaller than 1 << GLFix::precision otherwise the pixels can't be directly addressed
typedef struct TEXTURE { int width; int height; COLOR *bitmap; } TEXTURE;

class MATRIX {
public:
    MATRIX() { memset(data, 0, sizeof(data)); }
    GLFix data[4][4];
};

#define GL_COLOR_BUFFER_BIT 1<<0
#define GL_DEPTH_BUFFER_BIT 1<<1

enum GLDrawMode
{
    GL_TRIANGLES,
    GL_QUADS,
    GL_QUAD_STRIP //Not really tested
};

//Range [0-1]
typedef struct RGB { GLFix r; GLFix g; GLFix b; } RGB;

RGB rgbColor(COLOR c);
COLOR colorRGB(GLFix r, GLFix g, GLFix b);

//Invoke once before using any other functions
void nglInit();
void nglUninit();
//The buffer to render to
void nglSetBuffer(COLOR *screenBuf);
//Display the buffer
void nglDisplay();
void nglSetColor(COLOR c);
void nglRotateX(GLFix a);
void nglRotateY(GLFix a);
void nglRotateZ(GLFix a);
//Every vertex added will be appended to this buffer. It has to be allocated with enough capacity
void nglStartRecording(VERTEX *buffer);
//This returns the count of vertices added to the buffer
int nglStopRecording();
void nglAddVertices(VERTEX *buffer, int length);
void nglAddVertex(VERTEX *vertex);
void nglDrawTriangle(VERTEX *low, VERTEX *middle, VERTEX *high);
void nglDrawLine3D(VERTEX *v1, VERTEX *v2);

void glLoadIdentity();
void glBegin(GLDrawMode mode);
inline void glEnd() { }
void glClear(int buffers);
void glTranslatef(GLFix x, GLFix y, GLFix z);

void glBindTexture(TEXTURE *tex);
void glTexCoord2f(GLFix nu, GLFix nv);
void glColor3f(GLFix r, GLFix g, GLFix b);
void glVertex3f(GLFix x, GLFix y, GLFix z);
void glScale3f(GLFix x, GLFix y, GLFix z);
void glPushMatrix();
void glPopMatrix();

#endif
